@@ -1,11 +1,11 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /home/t-ishii/repository/pgpool/child.c,v 1.40 2004/11/12 13:59:12 t-ishii Exp $
+ * $Header: /cvsroot/pgpool/pgpool/child.c,v 1.1.1.1 2005/02/14 14:52:21 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003, 2004	Tatsuo Ishii
+ * Copyright (c) 2003-2005	Tatsuo Ishii
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -53,7 +53,7 @@ static void set_nonblock(int fd);
 static void unset_nonblock(int fd);
 #endif
 
-static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, int *is_timeout);
+static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *timeout);
 static StartupPacket *read_startup_packet(POOL_CONNECTION *cp);
 static int send_startup_packet(POOL_CONNECTION_POOL_SLOT *cp);
 static POOL_CONNECTION_POOL *connect_backend(StartupPacket *sp, POOL_CONNECTION *frontend);
@@ -79,7 +79,8 @@ void do_child(int unix_fd, int inet_fd)
 	struct timeval now;
 	struct timezone tz;
 	int child_idle_sec;
-	int connected;
+	struct timeval timeout;
+	static int connected;
 
 	pool_debug("I am %d", getpid());
 
@@ -113,15 +114,17 @@ void do_child(int unix_fd, int inet_fd)
 	}
 
 	child_idle_sec = 0;
-	connected = 0;
+
+	timeout.tv_sec = pool_config.child_life_time;
+	timeout.tv_usec = 0;
 
 	for (;;)
 	{
 		int connection_reuse = 1;
 		int ssl_request = 0;
 		StartupPacket *sp;
-		int is_timeout = 0;
 
+		/* pgpool stop request already sent? */
 		if (exit_request)
 		{
 			die(0);
@@ -131,13 +134,13 @@ void do_child(int unix_fd, int inet_fd)
 		idle = 1;
 
 		/* perform accept() */
-		frontend = do_accept(unix_fd, inet_fd, &is_timeout);
+		frontend = do_accept(unix_fd, inet_fd, &timeout);
 
-		if (connected && is_timeout)
+		if (frontend == NULL)
 		{
-			/* select() timeout */
-			child_idle_sec++;
-			if (pool_config.child_life_time != 0 && child_idle_sec >= pool_config.child_life_time)
+			/* check select() timeout */
+			if (connected && pool_config.child_life_time > 0 &&
+				timeout.tv_sec == 0 && timeout.tv_usec == 0)
 			{
 				pool_debug("child life %d seconds expired", pool_config.child_life_time);
 				send_frontend_exits();
@@ -146,12 +149,6 @@ void do_child(int unix_fd, int inet_fd)
 			continue;
 		}
 
-		else if (frontend == NULL)
-		{
-			/* accept() failed. return to the accept() loop */
-			continue;
-		}
-
 #ifdef NOT_USED
 		/* set frontend fd to blocking */
 		unset_nonblock(frontend->fd);
@@ -178,7 +175,7 @@ void do_child(int unix_fd, int inet_fd)
 		if (sp->major == 1234 && sp->minor == 5678)
 		{
 			cancel_request((CancelPacket *)sp->startup_packet, 0);
-			if (pool_config.replication_enabled)
+			if (DUAL_MODE)
 				cancel_request((CancelPacket *)sp->startup_packet, 1);
 			pool_close(frontend);
 			pool_free_startup_packet(sp);
@@ -273,8 +270,7 @@ void do_child(int unix_fd, int inet_fd)
 		if (pool_config.replication_strict == 0)
 			pool_enable_timeout();
 
-		/* mark this process has accepted connection */
-		connected++;
+		connected = 1;
 
 		/* query process loop */
 		for (;;)
@@ -289,8 +285,14 @@ void do_child(int unix_fd, int inet_fd)
 			{
 				/* client exits */
 				case POOL_END:
-					/* do not cache connection to template0, template1, regression */
-					if (!strcmp(sp->database, "template0") || !strcmp(sp->database, "template1") ||
+					/*
+					 * do not cache connection if:
+					 * pool_config.connection_cahe == 0 or
+					 * datase name is template0, template1, or regression
+					 */
+					if (pool_config.connection_cache == 0 ||
+						!strcmp(sp->database, "template0") ||
+						!strcmp(sp->database, "template1") ||
 						!strcmp(sp->database, "regression"))
 					{
 						pool_close(frontend);
@@ -335,7 +337,7 @@ void do_child(int unix_fd, int inet_fd)
 
 				/* not implemented yet */
 				case POOL_IDLE:
-					do_accept(unix_fd, inet_fd, NULL);
+					do_accept(unix_fd, inet_fd, &timeout);
 					pool_debug("accept while idle");
 					break;
 
@@ -347,6 +349,9 @@ void do_child(int unix_fd, int inet_fd)
 				break;
 		}
 
+		timeout.tv_sec = pool_config.child_life_time;
+		timeout.tv_usec = 0;
+
 	}
 	exit(0);
 }
@@ -403,9 +408,9 @@ static void unset_nonblock(int fd)
 #endif
 
 /*
-* perform accept() and returns new fd
+* perform accept() and return new fd
 */
-static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, int *is_timeout)
+static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *timeout)
 {
     fd_set	readmask;
     int fds;
@@ -421,14 +426,63 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, int *is_timeout)
 	static long atime;
 	static int cnt;
 #endif
-	struct timeval timeout = {1, 0};
+	struct timeval *timeoutval;
+	struct timeval tv1, tv2, tmback;
 
 	FD_ZERO(&readmask);
 	FD_SET(unix_fd, &readmask);
 	if (inet_fd)
 		FD_SET(inet_fd, &readmask);
 
-	fds = select(Max(unix_fd, inet_fd)+1, &readmask, NULL, NULL, &timeout);
+	if (timeout->tv_sec == 0 && timeout->tv_usec == 0)
+		timeoutval = NULL;
+	else
+	{
+		timeoutval = timeout;
+		tmback.tv_sec = timeout->tv_sec;
+		tmback.tv_usec = timeout->tv_usec;
+		gettimeofday(&tv1, NULL);
+
+#ifdef DEBUG
+		pool_log("before select = {%d, %d}", timeoutval->tv_sec, timeoutval->tv_usec);
+		pool_log("g:before select = {%d, %d}", tv1.tv_sec, tv1.tv_usec);
+#endif
+	}
+
+	fds = select(Max(unix_fd, inet_fd)+1, &readmask, NULL, NULL, timeoutval);
+
+	/*
+	 * following code fragment computes remaining timeout val in a
+	 * portable way. Linux does this automazically but other platforms do not.
+	 */
+	if (timeoutval)
+	{
+		gettimeofday(&tv2, NULL);
+
+		tmback.tv_usec -= tv2.tv_usec - tv1.tv_usec;
+		tmback.tv_sec -= tv2.tv_sec - tv1.tv_sec;
+
+		if (tmback.tv_usec < 0)
+		{
+			tmback.tv_sec--;
+			if (tmback.tv_sec < 0)
+			{
+				timeout->tv_sec = 0;
+				timeout->tv_usec = 0;
+			}
+			else
+			{
+				tmback.tv_usec += 1000000;
+				timeout->tv_sec = tmback.tv_sec;
+				timeout->tv_usec = tmback.tv_usec;
+			} 
+		}
+#ifdef DEBUG
+		pool_log("g:after select = {%d, %d}", tv2.tv_sec, tv2.tv_usec);
+		pool_log("after select = {%d, %d}", timeout->tv_sec, timeout->tv_usec);
+#endif
+	}
+
 	if (fds == -1)
 	{
 		if (errno == EAGAIN || errno == EINTR)
@@ -441,7 +495,6 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, int *is_timeout)
 	/* timeout */
 	if (fds == 0)
 	{
-		*is_timeout = 1;
 		return NULL;
 	}
 
@@ -735,7 +788,7 @@ static POOL_CONNECTION_POOL *connect_backend(StartupPacket *sp, POOL_CONNECTION
 	 */
 	backend->slots[0]->sp = sp;
 
-	if (pool_config.replication_enabled)
+	if (DUAL_MODE)
 	{
 		backend->slots[1]->con->isbackend = 1;
 		backend->slots[1]->con->issecondary_backend = 1;
@@ -755,7 +808,7 @@ static POOL_CONNECTION_POOL *connect_backend(StartupPacket *sp, POOL_CONNECTION
 	}
 
 	/* send startup packet */
-	if (pool_config.replication_enabled)
+	if (DUAL_MODE)
 	{
 		if (send_startup_packet(backend->slots[1]) < 0)
 		{
@@ -854,3 +907,100 @@ void pool_free_startup_packet(StartupPacket *sp)
 		free(sp);
 	}
 }
+
+/*
+ * check if we can connect to the backend
+ * returns 0 for ok. -1 for master down, -2 for secondary down.
+ */
+int health_check(void)
+{
+	int fd;
+
+	/* V2 startup packet */
+	typedef struct {
+		int len;		/* startup packet length */
+		StartupPacket_v2 sp;
+	} MySp;
+	MySp mysp;
+	char kind;
+
+	memset(&mysp, 0, sizeof(mysp));
+	mysp.len = htonl(296);
+	mysp.sp.protoVersion = htonl(PROTO_MAJOR_V2 << 16);
+	strcpy(mysp.sp.database, "template1");
+	strcpy(mysp.sp.user, pool_config.health_check_user);
+	*mysp.sp.options = '\0';
+	*mysp.sp.unused = '\0';
+	*mysp.sp.tty = '\0';
+
+	if (*pool_config.current_backend_host_name == '\0')
+		fd = connect_unix_domain_socket(0);
+	else
+		fd = connect_inet_domain_socket(0);
+
+	if (fd < 0)
+	{
+		pool_error("health check failed. master %s at port %d is down",
+				   pool_config.current_backend_host_name,
+				   pool_config.current_backend_port);
+		return -1;
+	}
+
+	if (write(fd, &mysp, sizeof(mysp)) < 0)
+	{
+		pool_error("health check failed during write. master %s at port %d is down",
+				   pool_config.current_backend_host_name,
+				   pool_config.current_backend_port);
+		return -1;
+	}
+
+	read(fd, &kind, 1);
+
+	if (write(fd, "X", 1) < 0)
+	{
+		pool_error("health check failed during write. master %s at port %d is down",
+				   pool_config.current_backend_host_name,
+				   pool_config.current_backend_port);
+		return -1;
+	}
+
+	close(fd);
+
+	if (!DUAL_MODE)
+		return 0;
+
+	if (*pool_config.secondary_backend_host_name == '\0')
+		fd = connect_unix_domain_socket(1);
+	else
+		fd = connect_inet_domain_socket(1);
+
+	if (fd < 0)
+	{
+		pool_error("health check failed. secondary %s at port %d is down",
+				   pool_config.secondary_backend_host_name,
+				   pool_config.secondary_backend_port);
+		return -2;
+	}
+
+	if (write(fd, &mysp, sizeof(mysp)) < 0)
+	{
+		pool_error("health check failed during write. master %s at port %d is down",
+				   pool_config.current_backend_host_name,
+				   pool_config.current_backend_port);
+		return -1;
+	}
+
+	read(fd, &kind, 1);
+
+	if (write(fd, "X", 1) < 0)
+	{
+		pool_error("health check failed during write. master %s at port %d is down",
+				   pool_config.current_backend_host_name,
+				   pool_config.current_backend_port);
+		return -1;
+	}
+
+	close(fd);
+
+	return 0;
+}
