@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool/main.c,v 1.2 2005/02/20 04:17:37 t-ishii Exp $
+ * $Header: /cvsroot/pgpool/pgpool/main.c,v 1.3 2005/02/25 06:38:16 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -295,10 +295,11 @@ int main(int argc, char **argv)
 			/*
 			 * do actual health check. trying to connect to the backend
 			 */
+			errno = 0;
 			health_check_timer_expired = 0;
 			sts = health_check();
 
-			if (health_check_timer_expired)
+			if (errno != EINTR || (errno == EINTR && health_check_timer_expired))
 			{
 				if (sts == -1)
 				{
