@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /home/t-ishii/repository/pgpool/pool_connection_pool.c,v 1.29 2005/01/30 13:04:06 t-ishii Exp $
+ * $Header: /cvsroot/pgpool/pgpool/pool_connection_pool.c,v 1.2 2005/03/12 12:56:39 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -194,7 +194,7 @@ void pool_connection_pool_timer(POOL_CONNECTION_POOL *backend)
 	POOL_CONNECTION_POOL *p = pool_connection_pool;
 	int i;
 
-	pool_debug("pool_connection_pool_timer: called");
+	pool_debug("pool_connection_pool_timer: set close time %d", time(NULL));
 
 	MASTER_CONNECTION(backend)->closetime = time(NULL);		/* set connection close time */
 
