@@ -1,11 +1,11 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /home/t-ishii/repository/pgpool/pool_process_query.c,v 1.37 2004/12/06 10:56:42 t-ishii Exp $
+ * $Header: /cvsroot/pgpool/pgpool/pool_process_query.c,v 1.3 2005/03/12 13:01:01 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003, 2004	Tatsuo Ishii
+ * Copyright (c) 2003-2005	Tatsuo Ishii
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -98,6 +98,9 @@ static void end_load_balance(POOL_CONNECTION_POOL *backend);
 
 static POOL_CONNECTION_POOL_SLOT *slots[MAX_CONNECTION_SLOTS];
 static int in_load_balance;
+static int master_slave_dml;
+static int replication_was_enabled;
+static int master_slave_was_enabled;
 
 POOL_STATUS pool_process_query(POOL_CONNECTION *frontend, 
 							   POOL_CONNECTION_POOL *backend,
@@ -131,7 +134,14 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 			st = reset_backend(backend, qcnt);
 
 			if (st < 0)		/* error? */
+			{
+				/* probably we don't need this, since caller will
+				 * close the connection to frontend after returning with POOL_END. But I
+				 * guess I would like to be a paranoid...
+				 */
+				frontend->no_forward = 0;
 				return POOL_END;
+			}
 
 			else if (st == 0)	/* no query issued? */
 			{
@@ -146,7 +156,7 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 				continue;
 			}
 
-			else if (st == 2)	/* no more qury */
+			else	/* no more query(st == 2) */
 			{
 				frontend->no_forward = 0;
 				return POOL_CONTINUE;
@@ -154,8 +164,8 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 
 		}
 
-		if ((!REPLICATION && MASTER(backend)->len == 0 && frontend->len == 0) ||
-			(REPLICATION && MASTER(backend)->len == 0 &&
+		if ((!DUAL_MODE && MASTER(backend)->len == 0 && frontend->len == 0) ||
+			(DUAL_MODE && MASTER(backend)->len == 0 &&
 			SECONDARY(backend)->len == 0
 			 && frontend->len == 0))
 		{
@@ -171,7 +181,7 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 			if (!connection_reuse)
 				FD_SET(frontend->fd, &readmask);
 			FD_SET(MASTER(backend)->fd, &readmask);
-			if (REPLICATION)
+			if (DUAL_MODE)
 				FD_SET(SECONDARY(backend)->fd, &readmask);
 			if (!connection_reuse)
 				FD_SET(frontend->fd, &exceptmask);
@@ -179,7 +189,7 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 
 			if (connection_reuse)
 			{
-				if (REPLICATION)
+				if (DUAL_MODE)
 					fds = select(Max(SECONDARY(backend)->fd, MASTER(backend)->fd) + 1,
 								 &readmask, &writemask, &exceptmask, NULL);
 				else
@@ -187,7 +197,7 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 			}
 			else
 			{
-				if (REPLICATION)
+				if (DUAL_MODE)
 					fds = select(Max(SECONDARY(backend)->fd,
 									 Max(frontend->fd, MASTER(backend)->fd)+1),
 								 &readmask, &writemask, &exceptmask, NULL);
@@ -216,7 +226,7 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 				pool_debug("read kind from backend %c", kind);
 			}
 
-			if (REPLICATION && FD_ISSET(SECONDARY(backend)->fd, &readmask))
+			if (DUAL_MODE && FD_ISSET(SECONDARY(backend)->fd, &readmask))
 			{
 				pool_read(SECONDARY(backend), &kind1, 1);
 				pool_debug("read kind from secondary backend %c", kind1);
@@ -245,17 +255,32 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 			if (MASTER(backend)->len > 0)
 			{
 				pool_read(MASTER(backend), &kind, 1);
-				if (REPLICATION)
+				if (DUAL_MODE)
 				{
 					pool_read(SECONDARY(backend), &kind1, 1);
 					if (kind == '\0' || kind != kind1)
 					{
+						int sts;
+
 						pool_error("pool_process_query: kind does not match between backends master(%c) secondary(%c)",
 								   kind, kind1);
 						pool_send_error_message(frontend, MAJOR(backend), "XX000", 
 												"kind mismatch between backends", "",
 												"check data consistency between master and secondary", __FILE__, __LINE__);
 
+						/* health check */
+						sts = health_check();
+						if (sts == -1)
+						{
+							notice_backend_error(1);
+							exit(1);
+						}
+						else if (sts == -2)
+						{
+							notice_backend_error(0);
+							exit(1);
+						}
+
 						if (pool_config.replication_stop_on_mismatch)
 							return POOL_FATAL;
 						else
@@ -275,7 +300,7 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 		}
 
 		/* this is the synchronous point */
-		if (REPLICATION)
+		if (DUAL_MODE)
 		{
 			if (kind == 0)
 			{
@@ -311,12 +336,27 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 
 			if (kind == '\0' || kind != kind1)
 			{
+				int sts;
+
 				pool_error("pool_process_query: kind does not match between backends master(%c) secondary(%c)",
 						   kind, kind1);
 				pool_send_error_message(frontend, MAJOR(backend), "XX000", 
 										"kind mismatch between backends", "",
 										"check data consistency between master and secondary", __FILE__, __LINE__);
 
+				/* health check */
+				sts = health_check();
+				if (sts == -1)
+				{
+					notice_backend_error(1);
+					exit(1);
+				}
+				else if (sts == -2)
+				{
+					notice_backend_error(0);
+					exit(1);
+				}
+
 				if (pool_config.replication_stop_on_mismatch)
 					return POOL_FATAL;
 				else
@@ -448,7 +488,7 @@ static POOL_STATUS Query(POOL_CONNECTION *frontend,
 	int len;
 	static char *sq = "show pool_status";
 
-	if (query == NULL)
+	if (query == NULL)	/* need to read query from frontend? */
 	{
 		/* read actual query */
 		if (MAJOR(backend) == PROTO_MAJOR_V3)
@@ -483,6 +523,12 @@ static POOL_STATUS Query(POOL_CONNECTION *frontend,
 	/* load balance trick */
 	if (load_balance_enabled(backend, string))
 		start_load_balance(backend);
+	else if (MASTER_SLAVE)
+	{
+		master_slave_was_enabled = 1;
+		MASTER_SLAVE = 0;
+		master_slave_dml = 1;
+	}
 
 	/* forward the query to the backend */
 	pool_write(MASTER(backend), "Q", 1);
@@ -499,11 +545,36 @@ static POOL_STATUS Query(POOL_CONNECTION *frontend,
 	}
 
 	if (REPLICATION)
+#ifdef NOT_USED
+		(frontend == NULL &&  )		/* we assume that frontend == NULL
+								 * means that Query() is called from
+								 * reset_backend(). In master/slave
+								 * mode AND load balance is enabled,
+								 * we need to send reset queries to
+								 * secondary as well.
+								 */
+#endif
 	{
-		/* in "strict mode" we need to wait for master completing the query */
-		if (pool_config.replication_strict || STRICT_MODE(string))
+		/*
+		 * in "strict mode" we need to wait for master completing the query.
+		 * note that this is applied to non SELECT queries only.
+		 */
+		if (strncasecmp("SELECT", string, 6) &&
+			(pool_config.replication_strict || STRICT_MODE(string)))
+		{
 			if (synchronize(MASTER(backend)))
 				return POOL_END;
+		}
+
+#define SEQUENCE_DEBUG
+#ifdef SEQUENCE_DEBUG
+		if (!strncmp(string, "/*SLEEP*/", 9))
+		{
+			pool_debug("start sleeping");
+			sleep(20);
+			pool_debug("end sleeping");
+		}
+#endif
 
 		pool_write(SECONDARY(backend), "Q", 1);
 		if (MAJOR(backend) == PROTO_MAJOR_V3)
@@ -568,6 +639,13 @@ static POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 	if (in_load_balance)
 		end_load_balance(backend);
 
+	if (master_slave_dml)
+	{
+		MASTER_SLAVE = 1;
+		master_slave_was_enabled = 0;
+		master_slave_dml = 0;
+	}
+
 	return ProcessFrontendResponse(frontend, backend);
 }
 
@@ -590,7 +668,7 @@ static POOL_STATUS CompleteCommandResponse(POOL_CONNECTION *frontend,
 
 		if (len != len1)
 		{
-			pool_log("Complete Command Response: message length does not match between master(%d \"%s\",) and secondary(%d \"%s\",)",
+			pool_debug("Complete Command Response: message length does not match between master(%d \"%s\",) and secondary(%d \"%s\",)",
 					 len, string, len1, string1);
 		}
 	}
@@ -665,7 +743,7 @@ static int RowDescription(POOL_CONNECTION *frontend,
 			/* we do not regard oid mismatch as fatal */
 			if (oid != oid1)
 			{
-				pool_log("RowDescription: field oid deos not match between backends master(%d) and secondary(%d)",
+				pool_debug("RowDescription: field oid deos not match between backends master(%d) and secondary(%d)",
 						 ntohl(oid), ntohl(oid1));
 			}
 		}
@@ -678,7 +756,7 @@ static int RowDescription(POOL_CONNECTION *frontend,
 			pool_read(SECONDARY(backend), &size1, sizeof(short));
 			if (size1 != size1)
 			{
-				pool_log("RowDescription: field size deos not match between backends master(%d) and secondary(%d)",
+				pool_error("RowDescription: field size deos not match between backends master(%d) and secondary(%d)",
 						 ntohs(size), ntohs(size1));
 				return POOL_FATAL;
 			}
@@ -693,7 +771,7 @@ static int RowDescription(POOL_CONNECTION *frontend,
 			pool_read(SECONDARY(backend), &mod1, sizeof(int));
 			if (mod != mod1)
 			{
-				pool_log("RowDescription: modifier deos not match between backends master(%d) and secondary(%d)",
+				pool_debug("RowDescription: modifier deos not match between backends master(%d) and secondary(%d)",
 						 ntohl(mod), ntohl(mod1));
 			}
 		}
@@ -739,7 +817,7 @@ static POOL_STATUS AsciiRow(POOL_CONNECTION *frontend,
 			   this as a fatal error. However in the real world
 			   we'd better to adapt this situation. Just throw a
 			   log... */
-			pool_log("AsciiRow: NULLMAP differ between master and secondary");
+			pool_debug("AsciiRow: NULLMAP differ between master and secondary");
 		}
 	}
 
@@ -770,7 +848,7 @@ static POOL_STATUS AsciiRow(POOL_CONNECTION *frontend,
 				return POOL_END;
 
 			if (size != size1)
-				pool_log("AsciiRow: %d th field size does not match between master(%d) and secondary(%d)",
+				pool_debug("AsciiRow: %d th field size does not match between master(%d) and secondary(%d)",
 						 i, ntohl(size), ntohl(size1));
 			size1 = ntohl(size1) - 4;
 		}
@@ -847,7 +925,7 @@ static POOL_STATUS BinaryRow(POOL_CONNECTION *frontend,
 			   this as a fatal error. However in the real world
 			   we'd better to adapt this situation. Just throw a
 			   log... */
-			pool_log("BinaryRow: NULLMAP differ between master and secondary");
+			pool_debug("BinaryRow: NULLMAP differ between master and secondary");
 		}
 	}
 
@@ -878,7 +956,7 @@ static POOL_STATUS BinaryRow(POOL_CONNECTION *frontend,
 				return POOL_END;
 
 			if (size != size1)
-				pool_log("BinaryRow: %d th field size does not match between master(%d) and secondary(%d)",
+				pool_debug("BinaryRow: %d th field size does not match between master(%d) and secondary(%d)",
 						 i, ntohl(size), ntohl(size1));
 			size1 = ntohl(size1) - 4;
 		}
@@ -1328,7 +1406,7 @@ static POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 
 	if (pool_read(frontend, &fkind, 1) < 0)
 	{
-		pool_error("ProcessFrontendResponse: failed to read kind");
+		pool_error("ProcessFrontendResponse: failed to read kind from frontend. fronend abnormally exited");
 		return POOL_END;
 	}
 
@@ -1425,7 +1503,7 @@ int pool_check_fd(POOL_CONNECTION *cp, int notimeout)
 		else
 		{
 			timeout.tv_sec = pool_config.replication_timeout / 1000;
-			timeout.tv_usec = pool_config.replication_timeout - (timeout.tv_sec * 1000);
+			timeout.tv_usec = (pool_config.replication_timeout - (timeout.tv_sec * 1000))*1000;
 			tp = &timeout;
 		}
 
@@ -1450,7 +1528,7 @@ int pool_check_fd(POOL_CONNECTION *cp, int notimeout)
 		{
 			pool_error("pool_check_fd: data is not ready tp->tv_sec %d tp->tp_usec %d", 
 					   pool_config.replication_timeout / 1000,
-					   pool_config.replication_timeout - (timeout.tv_sec * 1000));
+					   (pool_config.replication_timeout - (timeout.tv_sec * 1000))*1000);
 			break;
 		}
 		return 0;
@@ -1467,7 +1545,7 @@ static void process_reporting(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *b
 	static short fsize = -1;
 	static int mod = 0;
 	short n;
-	int i;
+	int i, j;
 	short s;
 	int len;
 	short colnum;
@@ -1593,6 +1671,49 @@ static void process_reporting(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *b
 	status[i].desc = "stop replication mode on fatal error";
 	i++;
 
+	status[i].name = "reset_query_list";
+	*(status[i].value) = '\0';
+	for (j=0;j<pool_config.num_reset_queries;j++)
+	{
+		int len;
+		len = MAXVALLEN - strlen(status[i].value);
+		strncat(status[i].value, pool_config.reset_query_list[j], len);
+		len = MAXVALLEN - strlen(status[i].value);
+		strncat(status[i].value, ";", len);
+	}
+	status[i]. desc = "queries issued at the end of session";
+	i++;
+
+	status[i].name = "print_timestamp";
+	snprintf(status[i].value, MAXVALLEN, "%d", pool_config.print_timestamp);
+	status[i].desc = "if true print time stamp to each log line";
+	i++;
+
+	status[i].name = "master_slave_mode";
+	snprintf(status[i].value, MAXVALLEN, "%d", pool_config.master_slave_mode);
+	status[i].desc = "if true, operate in master/slave mode";
+	i++;
+		 
+	status[i].name = "connection_cache";
+	snprintf(status[i].value, MAXVALLEN, "%d", pool_config.connection_cache);
+	status[i].desc = "if true, cache connection pool";
+	i++;
+
+	status[i].name = "health_check_timeout";
+	snprintf(status[i].value, MAXVALLEN, "%d", pool_config.health_check_timeout);
+	status[i].desc = "health check timeout";
+	i++;
+
+	status[i].name = "health_check_period";
+	snprintf(status[i].value, MAXVALLEN, "%d", pool_config.health_check_period);
+	status[i].desc = "health check period";
+	i++;
+
+	status[i].name = "health_check_user";
+	snprintf(status[i].value, MAXVALLEN, "%s", pool_config.health_check_user);
+	status[i].desc = "health check user";
+	i++;
+
 	status[i].name = "current_backend_host_name";
 	snprintf(status[i].value, MAXVALLEN, "%s", pool_config.current_backend_host_name);
 	status[i].desc = "current master host name";
@@ -1608,6 +1729,16 @@ static void process_reporting(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *b
 	status[i].desc = "non 0 if actually operating in replication mode";
 	i++;
 
+	status[i].name = "master_slave_enabled";
+	snprintf(status[i].value, MAXVALLEN, "%d", pool_config.master_slave_enabled);
+	status[i].desc = "non 0 if actually operating in master/slave";
+	i++;
+
+	status[i].name = "num_reset_queries";
+	snprintf(status[i].value, MAXVALLEN, "%d", pool_config.num_reset_queries);
+	status[i].desc = "number of queries in reset_query_list";
+	i++;
+
 	nrows = i;
 
 	if (MAJOR(backend) == PROTO_MAJOR_V2)
@@ -1764,7 +1895,7 @@ void pool_send_frontend_exits(POOL_CONNECTION_POOL *backend)
 	 */
 	fflush(MASTER(backend)->write_fd);
 
-	if (REPLICATION)
+	if (DUAL_MODE)
 	{
 		pool_write(SECONDARY(backend), "X", 1);
 		if (MAJOR(backend) == PROTO_MAJOR_V3)
@@ -1807,7 +1938,7 @@ POOL_STATUS SimpleForwardToFrontend(char kind, POOL_CONNECTION *frontend, POOL_C
 
 		if (len != len1)
 		{
-			pool_log("SimpleForwardToFrontend: length does not match between backends master(%d) secondary(%d) kind:(%c)",
+			pool_debug("SimpleForwardToFrontend: length does not match between backends master(%d) secondary(%d) kind:(%c)",
 					 ntohl(len), ntohl(len1), kind);
 		}
 	}
@@ -1907,7 +2038,7 @@ POOL_STATUS ParameterStatus(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *bac
 	pool_param_debug_print(&MASTER(backend)->params);
 #endif
 
-	if (REPLICATION)
+	if (DUAL_MODE)
 		if (pool_read2(SECONDARY(backend), len) == NULL)
 			return POOL_END;
 
@@ -1921,19 +2052,15 @@ POOL_STATUS ParameterStatus(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *bac
  */
 static int reset_backend(POOL_CONNECTION_POOL *backend, int qcnt)
 {
-#ifdef NO_RESET_ALL
-	static char *queries[] = {"ABORT"};
-#else
-	static char *queries[] = {"ABORT", "RESET ALL"};
-#endif
-
 	char *query;
-	int qn = sizeof(queries)/sizeof(char *);
+	int qn;
+
+	qn = pool_config.num_reset_queries;
 
 	if (qcnt >= qn)
 		return 2;
 
-	query = queries[qcnt];
+	query = pool_config.reset_query_list[qcnt];
 
 	/* if transaction state is idle, we don't need to issue ABORT */
 	if (TSTATE(backend) == 'I' && !strcmp("ABORT", query))
@@ -1951,7 +2078,7 @@ static int reset_backend(POOL_CONNECTION_POOL *backend, int qcnt)
 static int load_balance_enabled(POOL_CONNECTION_POOL *backend, char *sql)
 {
 	if (pool_config.load_balance_mode &&
-		REPLICATION &&
+		DUAL_MODE &&
 		MAJOR(backend) == PROTO_MAJOR_V3 &&
 		TSTATE(backend) == 'I' &&
 		!strncasecmp(sql, "SELECT", 6))
@@ -1974,7 +2101,13 @@ static void start_load_balance(POOL_CONNECTION_POOL *backend)
 	}
 
 	/* temporarily turn off replication mode */
+	if (REPLICATION)
+		replication_was_enabled = 1;
+	if (MASTER_SLAVE)
+		master_slave_was_enabled = 1;
+
 	REPLICATION = 0;
+	MASTER_SLAVE = 0;
 
 	/* choose a master in random manner with weight */
 	master = (random() <= weight_master)?0:1;
@@ -1999,9 +2132,13 @@ static void end_load_balance(POOL_CONNECTION_POOL *backend)
 	}
 
 	/* turn on replication mode */
-	REPLICATION = 1;
+	REPLICATION = replication_was_enabled;
+	MASTER_SLAVE = master_slave_was_enabled;
 
+	replication_was_enabled = 0;
+	master_slave_was_enabled = 0;
 	in_load_balance = 0;
+
 	pool_debug("end_load_balance: end load balance mode");
 }
 
